from time import time, gmtime
import dataclasses as dtc
from typing import Optional

import ffmpeg
import h5mapper
import soundfile as sf
import numpy as np
import os

import torch
from pytorch_lightning import LightningModule
from pytorch_lightning.utilities import rank_zero_only
from pytorch_lightning.loggers import LightningLoggerBase
from pytorch_lightning.loggers.base import rank_zero_experiment

__all__ = [
    "LoggingHooks",
    "LossLogger",
    'AudioLogger',
    'convert_to_mp3'
]


class LoggingHooks(LightningModule):

    def on_epoch_start(self):
        # convenience to have printing of the loss at the end of the epoch
        self._ep_metrics = {}
        self._batch_count = {}

    def log_output(self, out):
        for metric, val in out.items():
            if torch.isnan(val.detach()):
                raise KeyboardInterrupt(f"metric {metric} is nan")
            if metric not in self._ep_metrics:
                self._ep_metrics.setdefault(metric, val.detach())
                self._batch_count[metric] = 0
            else:
                self._ep_metrics[metric] += val.detach()
            self._batch_count[metric] += 1
        return out

    def training_step_end(self, out):
        return self.log_output(out)

    def validation_step_end(self, out):
        return self.log_output(out)

    def _flush_ep_metrics(self):
        to_print = "Epoch %i " % self.current_epoch
        to_log = {}
        for k, v in self._ep_metrics.items():
            to_print += "- %s : %.4f " % (k, v / self._batch_count[k])
            to_log[k] = (v / self._batch_count[k]).item()
        self.print(to_print)
        if getattr(self, "logger", None) is not None:
            self.logger.log_metrics(to_log, self.current_epoch)

    def on_train_epoch_end(self, *args):
        if self.trainer.val_dataloaders is not None and any(self.trainer.val_dataloaders):
            # wait until validation end
            return
        else:
            self._flush_ep_metrics()

    def on_validation_epoch_end(self, *args):
        self._flush_ep_metrics()

    def on_fit_start(self):
        self._ep_time = time()

    def on_fit_end(self):
        duration = time() - self._ep_time
        total_time = gmtime(duration)
        self.print("Training finished after "
                   "{0} days {1} hours {2} mins {3} seconds".format(total_time[2] - 1, total_time[3],
                                                                    total_time[4], total_time[5]))
        if getattr(self, "logger", None) is not None:
            # accumulate if we resumed
            duration = self.hparams.get("training_time_sec", 0) + duration
            self.hparams.update({"training_time_sec": duration})
            self.logger.log_hyperparams(self.hparams)


class LossLogger(LightningLoggerBase):

    def __init__(self, logs_file):
        super(LossLogger, self).__init__()
        self.logs_file = logs_file

    @property
    def name(self):
        return 'LossLogger'

    @property
    @rank_zero_experiment
    def experiment(self):
        # Return the experiment object associated with this logger.
        return None

    @property
    def version(self):
        # Return the experiment version, int or str.
        return '0.1'

    @rank_zero_only
    def log_hyperparams(self, params):
        # params is an argparse.Namespace
        # your code to record hyperparameters goes here
        pass

    @rank_zero_only
    def log_metrics(self, metrics, step):
        # metrics is a dictionary of metric names and values
        # your code to record metrics goes here
        self.logs_file.add(str(step), {k: np.array([v]) for k, v in metrics.items()})

    @rank_zero_only
    def save(self):
        # Optional. Any code necessary to save logger data goes here
        # If you implement this, remember to call `super().save()`
        # at the start of the method (important for aggregation of metrics)
        super().save()
        self.logs_file.flush()

    @rank_zero_only
    def finalize(self, status):
        # Optional. Any code that needs to be run after training
        # finishes goes here
        pass


@dtc.dataclass
class AudioLogger:
    sr: int = 16000
    hop_length: int = 512
    filename_template: Optional[str] = None
    target_dir: Optional[str] = None
    id_template: Optional[str] = None
    proxy_template: Optional[str] = None
    target_bank: Optional[h5mapper.TypedFile] = None

    @staticmethod
    def format_template(template, **parameters):
        exec(f"out = f'{template}'", {}, parameters)
        return parameters["out"]

    def log_mp3(self, audio, **params):
        filename = self.format_template(self.filename_template, **params)
        if '.wav' not in filename:
            filename += '.wav'
        if self.target_dir:
            os.makedirs(self.target_dir, exist_ok=True)
            if self.target_dir not in filename:
                filename = os.path.join(self.target_dir, filename)
        if isinstance(audio, torch.Tensor):
            audio = audio.squeeze().detach().cpu().numpy()
        sf.write(filename, audio, self.sr, 'PCM_24')
        convert_to_mp3(filename)

    def log_h5(self, audio, **params):
        if isinstance(audio, torch.Tensor):
            audio = audio.squeeze().detach().cpu().numpy()
        src = self.format_template(self.id_template, **params)
        proxy = self.format_template(self.proxy_template, **params)
        self.target_bank.add(src, {proxy: audio})
        self.target_bank.flush()

    def write(self, audio, **params):
        if self.filename_template and self.target_dir:
            self.log_mp3(audio, **params)
        if self.id_template and self.proxy_template and self.target_bank:
            self.log_h5(audio, **params)


def convert_to_mp3(file_path, exists_ok=True):
    if exists_ok and os.path.isfile(os.path.splitext(file_path)[0] + ".mp3"):
        os.remove(os.path.splitext(file_path)[0] + ".mp3")
    stream = ffmpeg.input(file_path)
    stream.output(os.path.splitext(file_path)[0] + ".mp3").run(quiet=True)
    os.remove(file_path)


import numpy as np
from scipy.sparse.csgraph import connected_components
from sklearn.metrics import pairwise_distances as pwd
from sklearn.neighbors import KNeighborsTransformer, NearestNeighbors
import sklearn.cluster as C
import torch
import torch.nn as nn
from joblib import Parallel, delayed
import click
import os
import soundfile as sf
import shutil
from functools import partial

from ..modules import angular_distance


class QCluster:

    def __init__(self, qe=.5, n_neighbs=None, k=None, metric="euclidean"):
        self.qe = qe
        self.n_neighbs = n_neighbs
        self.k = k
        self.metric = metric
        self.is_core_ = None
        self.labels_ = None
        self.K_ = None

    def fit(self, x):
        N = x.shape[0]
        if self.n_neighbs is None:
            self.n_neighbs = int(np.sqrt(N))
        if self.k is None:
            self.k = int(self.qe * self.n_neighbs)

        D = pwd(x, x, metric=self.metric)
        Dk = D.copy()
        Dk[Dk == 0] = np.inf

        kn = KNeighborsTransformer(mode="distance", n_neighbors=self.n_neighbs,
                                   metric="precomputed")
        adj = kn.fit_transform(D)
        adj_sub = kn.kneighbors_graph(None, self.k, mode="distance")
        Kx = (0 < adj.A).sum(axis=0).reshape(-1)
        is_core = Kx >= np.quantile(Kx, self.qe)

        adj_cores = (adj_sub.A.astype(np.bool)) & (is_core & is_core[:, None])
        cores_idx = is_core.nonzero()[0]
        print(f"found {cores_idx.shape[0]} cores")
        for i in range(adj.shape[0]):
            if not np.any(adj_cores[i] & is_core):
                nearest_core = Dk[i, is_core].argmin()
                nearest_core = cores_idx[nearest_core]
                adj_cores[i, nearest_core] = True

        K, labels = connected_components(adj_cores)
        self.K_, self.labels_, self.is_core_ = K, labels, is_core
        return self


class GCluster:
    def __init__(self,
                 n_means,
                 n_iter=128,
                 lr=0.025,
                 betas=(0.05, 0.05),
                 metric="cosine",
                 eps=1e-6):
        self.n_means = n_means
        self.n_iter = n_iter
        self.lr = lr
        self.betas = betas
        self.metric = metric
        self.d_func = dict(euclidean=torch.cdist,
                           cosine=partial(angular_distance, eps=eps))[metric]
        self.K_ = None
        self.labels_ = None
        self.losses_ = None

    def fit(self, x):
        X = torch.from_numpy(x)
        H = nn.Parameter(X[torch.randint(0, X.shape[0], (self.n_means,))].clone())
        opt = torch.optim.Adam([H], lr=self.lr, betas=self.betas)
        losses = []
        for _ in range(self.n_iter):
            opt.zero_grad()
            L = self.d_func(H, X).mean()
            L = L - .5 * self.d_func(H, H).mean()
            L.backward()
            opt.step()
            losses += [L.item()]
        x = X.detach().cpu().numpy()
        h = H.detach().cpu().numpy()
        DXH = pwd(h, x, self.metric)
        hi, xi = np.unravel_index(DXH.argsort(None), DXH.shape)
        labels = np.zeros(x.shape[0], dtype=np.int)
        got = set()
        for label, i in zip(hi.flat[:], xi.flat[:]):
            if i not in got:
                labels[i] = label
                got.add(i)
        self.losses_ = losses
        self.K_ = self.n_means
        self.labels_ = labels
        return self


class HCluster:
    def __init__(self, max_iter=32, metric="cosine"):
        self.max_iter = max_iter
        self.metric = metric
        self.K_ = None
        self.labels_ = None

    def fit(self, x):
        Da = pwd(x, x, self.metric)
        xa = x.copy()
        Da[Da == 0] = np.inf
        LBS = np.zeros((x.shape[0], self.max_iter), dtype=np.int)

        for i in range(self.max_iter):
            Adj_nearest = np.zeros_like(Da, dtype=np.bool)
            nearest = Da.argmin(axis=1)
            Adj_nearest[np.arange(Da.shape[0]), nearest] = True

            K, labels = connected_components(Adj_nearest)
            if i == 0:
                LBS[:, 0] = labels
            else:
                LBS[:, i] = np.r_[[labels[LBS[n, i - 1]] for n in range(x.shape[0])]]

            xa = np.stack([xa[labels == k].mean(axis=0) for k in range(K)])
            Da = pwd(xa, xa, metric=self.metric)
            Da[Da == 0] = np.inf
            if K == 1:
                LBS = LBS[:, :i + 1]
                self.K_ = i + 1
                break

        self.labels_ = LBS
        return self


def distance_matrices(X, metric="euclidean", n_neighbors=1, radius=1e-3):
    Dx = pwd(X, X, metric=metric)
    NN = NearestNeighbors(n_neighbors=n_neighbors, radius=radius, metric="precomputed")
    NN.fit(Dx)
    Kx = NN.kneighbors_graph(n_neighbors=n_neighbors, mode='connectivity')
    Rx = NN.radius_neighbors_graph(radius=radius, mode='connectivity')
    return Dx, Kx, Rx


class ArgMax(object):
    def __init__(self):
        self.labels_ = None
        self.K_ = None

    def fit(self, X):
        maxes = np.argmax(X, axis=1)
        uniques, self.labels_ = np.unique(maxes, return_inverse=True)
        self.K_ = len(uniques)
        return self


def sk_cluster(X, Dx=None, n_clusters=128, metric="euclidean", estimator="argmax"):
    estimators = {

        "argmax": ArgMax(),

        "kmeans": C.KMeans(n_clusters=n_clusters,
                           n_init=4,
                           max_iter=200,
                           ),

        "spectral": C.SpectralClustering(n_clusters=n_clusters,
                                         affinity="nearest_neighbors",
                                         n_neighbors=32,
                                         assign_labels="discretize",
                                         ),

        "agglo_ward": C.AgglomerativeClustering(
            n_clusters=n_clusters,
            affinity="euclidean",
            compute_full_tree='auto',
            linkage='ward',
            distance_threshold=None, ),

        "agglo_single": C.AgglomerativeClustering(
            n_clusters=n_clusters,
            affinity="precomputed",
            compute_full_tree='auto',
            linkage='single',
            distance_threshold=None, ),

        "agglo_complete": C.AgglomerativeClustering(
            n_clusters=n_clusters,
            affinity="precomputed",
            compute_full_tree='auto',
            linkage='complete',
            distance_threshold=None, )

    }

    needs_distances = estimator in {"agglo_single", "agglo_complete"}

    if needs_distances:
        if Dx is None:
            Dx, _, _ = distance_matrices(X, metric=metric)
        X_ = Dx
    else:
        X_ = X

    cls = estimators[estimator]
    cls.fit(X_)

    return cls


def etl(input_file, sr, n_fft, hop_length):
    """STFT Extract-Transform-Load"""
    from mimikit import FileToSignal, MagSpec

    y = FileToSignal(sr)(input_file)
    S = MagSpec(n_fft, hop_length)(y)
    return y, S


def export(S, target_path, sr, n_fft, hop_length):
    from mimikit import GLA
    gla = GLA(n_fft, hop_length)
    if S.shape[0] <= 1:
        print("!!!!!", target_path)
        return
    # if S is too long, joblib sends it as np.memmap
    # which messes with gla()...
    y = gla(np.asarray(S))
    sf.write(f"{target_path}.wav", y, sr, 'PCM_24')
    return


@click.command()
@click.option("-f", "--input-file", help="file to be segmented")
@click.option("--sr", "-r", default=22050, help="the sample rate used for loading the file (default=22050)")
@click.option("--n-fft", "-d", default=2048, help="size of the fft (default=2048)")
@click.option("--hop-length", "-h", default=512, help="hop length of the stft (default=512)")
@click.option("--n-neighbs", "-n", default=None, type=int,
              help="size of the neighborhood used to estimate point-density")
@click.option("--qe", "-q", default=0.5, type=float,
              help="proportion of edge-points")
@click.option("--k", "-k", default=1, type=int,
              help="proportion of connecting core points")
def cluster(input_file: str,
            sr: int = 22050,
            n_fft: int = 2048,
            hop_length: int = 512,
            n_neighbs: int = None,
            qe: float = 0.5,
            k: int = 1
            ):
    y, S = etl(input_file, sr, n_fft, hop_length)
    est = QCluster(qe=qe, n_neighbs=n_neighbs, k=k, metric="euclidean")
    est.fit(S)

    clusters = [S[est.labels_ == i] for i in range(est.K_)]
    items, counts = np.unique(est.labels_, return_counts=True)
    srtd = (-counts).argsort()
    items, counts = items[srtd[:10]], counts[srtd[:10]]
    distrib_str = "\n    ".join([f"{d}  :  {c}"
                                 for d, c in zip(items, counts)]) + \
                  "\n    [  ...  ]\n"
    print(f"""
Data shape : {S.shape}
Parameters : n_neighbs={est.n_neighbs} ; k={est.k} ; qe={qe}
========>   found {est.K_} clusters
    """ + distrib_str
          )

    target_dir = os.path.splitext(input_file)[0]
    if os.path.exists(target_dir):
        shutil.rmtree(target_dir)
    os.makedirs(target_dir, exist_ok=True)
    Parallel(backend='multiprocessing') \
        (delayed(export)(s, f"{target_dir}/c{i}", sr, n_fft, hop_length)
         for i, s in enumerate(clusters))
    return



import torch.nn as nn
import numpy as np
from functools import partial

__all__ = [
    'LinearResampler',
    'Conv1dResampler',
    'LSTMResampler'
]


class LinearResampler(nn.Module):
    def __init__(self, in_d, t_factor, d_factor, **kwargs):
        super().__init__()

        self.fc = nn.Linear(in_d, int(in_d * t_factor * d_factor), **kwargs)
        self.tf, self.df = t_factor, d_factor

    def forward(self, x):
        B, T, D = x.size()
        x = self.fc(x)
        return x.reshape(B, int(T * self.tf), int(D * self.df))


class Conv1dResampler(nn.Module):

    def __init__(self, in_d, t_factor, d_factor, **kwargs):
        super().__init__()
        mod = nn.Conv1d if t_factor <= 1 else partial(nn.ConvTranspose1d, stride=t_factor)
        self.cv = mod(in_d, int(in_d * d_factor),
                      t_factor if t_factor >= 1 else int(1 / t_factor),
                      **kwargs)
        self.tf, self.df = t_factor, d_factor

    def forward(self, x):
        if len(x.size()) > 3:
            x = x.view(x.size(0), np.prod(x.shape[1:-1]), x.size(-1))
        B, T, D = x.size()
        if self.tf <= 1:
            x = x.view(-1, int(1 / self.tf), D).transpose(1, 2)
            x = self.cv(x).squeeze(-1).reshape(B, int(D * self.df), -1)
        else:
            x = x.transpose(1, 2)
            x = self.cv(x)
        return x.transpose(1, 2)


class LSTMResampler(nn.Module):

    def __init__(self, in_d, t_factor, d_factor, **kwargs):
        super().__init__()
        self.lstm = nn.LSTM(in_d, int(in_d * t_factor * d_factor), batch_first=True, **kwargs)
        self.tf, self.df = t_factor, d_factor

    def forward(self, x, hidden=None):
        B, T, D = x.size()
        x, hidden = self.lstm(x, hidden)
        return x.reshape(B, int(T * self.tf), int(D * self.df)), hidden


import torch
import torch.nn as nn
from typing import Optional
import dataclasses as dtc

from mimikit.modules.misc import Abs
from pytorch_lightning.utilities import AttributeDict

from ..networks.parametrized_gaussian import ParametrizedGaussian

__all__ = [
    'EncoderLSTM',
    'DecoderLSTM',
    'Seq2SeqLSTMNetwork',
    'MultiSeq2SeqLSTM'
]


@dtc.dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class EncoderLSTM(nn.Module):
    input_d: int = 512
    model_dim: int = 512
    num_layers: int = 1
    n_lstm: Optional[int] = 1
    bottleneck: Optional[str] = "add"
    n_fc: Optional[int] = 1
    bias: Optional[bool] = False
    weight_norm: Optional[bool] = False
    hop: int = 8
    with_tbptt: bool = False

    def __post_init__(self):
        nn.Module.__init__(self)
        self.lstms = nn.ModuleList([
            nn.LSTM(self.input_d if i == 0 else self.model_dim,
                    self.model_dim if self.bottleneck == "add" else self.model_dim // 2,
                    bias=self.bias,
                    num_layers=self.num_layers,
                    batch_first=True, bidirectional=True)
            for i in range(self.n_lstm)
        ])
        self.fc = nn.Sequential(
            *[nn.Sequential(nn.Linear(self.model_dim, self.model_dim), nn.Tanh()) for _ in range(self.n_fc - 1)],
            nn.Linear(self.model_dim, self.model_dim, bias=False),  # NO ACTIVATION !
        )
        if self.weight_norm:
            for lstm in self.lstms:
                for name, p in dict(lstm.named_parameters()).items():
                    if "weight" in name:
                        torch.nn.utils.weight_norm(lstm, name)
        self.hidden = None

    def forward(self, x):
        ht, ct = None, None
        hidden = self.reset_hidden(x, self.hidden)
        for i, lstm in enumerate(self.lstms):
            out, (ht, ct) = lstm(x, hidden)
            # sum forward and backward nets
            out = out.view(*out.size()[:-1], self.model_dim, 2).sum(dim=-1)
            # take residuals AFTER the first lstm
            x = out if i == 0 else x + out
            if self.with_tbptt:
                self.hidden = ht, ct
        states = self.first_and_last_states(x)
        out = self.fc(states)
        return out, (ht, ct)

    def first_and_last_states(self, sequence):
        rg = torch.arange(sequence.size(1) // self.hop)
        first_states = sequence[:, rg * self.hop, :]
        last_states = sequence[:, (rg + 1) * self.hop - 1, :]
        if self.bottleneck == "add":
            return first_states + last_states
        else:
            return torch.cat((first_states, last_states), dim=-1)

    def reset_hidden(self, x, h):
        if h is None or x.size(0) != h[0].size(1):
            B = x.size(0)
            h0 = torch.zeros(self.num_layers * 2, B, self.model_dim).to(x.device)
            c0 = torch.zeros(self.num_layers * 2, B, self.model_dim).to(x.device)
            return h0, c0
        else:
            return tuple(h_.detach() for h_ in h)


@dtc.dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class DecoderLSTM(nn.Module):
    model_dim: int = 512
    num_layers: int = 1
    bottleneck: Optional[str] = "add"
    bias: Optional[bool] = False
    weight_norm: Optional[tuple] = (False, False)

    def __post_init__(self):
        nn.Module.__init__(self)
        self.lstm1 = nn.LSTM(self.model_dim, self.model_dim if self.bottleneck == "add" else self.model_dim // 2,
                             bias=self.bias,
                             num_layers=self.num_layers, batch_first=True, bidirectional=True)
        self.lstm2 = nn.LSTM(self.model_dim, self.model_dim if self.bottleneck == "add" else self.model_dim // 2,
                             bias=self.bias,
                             num_layers=self.num_layers, batch_first=True, bidirectional=True)
        for lstm, wn in zip([self.lstm1, self.lstm2], self.weight_norm):
            if wn:
                for name, p in dict(lstm.named_parameters()).items():
                    if "weight" in name:
                        torch.nn.utils.weight_norm(lstm, name)

    def forward(self, x, hidden, cells):
        if hidden is None or cells is None:
            output, (_, _) = self.lstm1(x)
        else:
            # ALL decoders get hidden states from encoder
            output, (_, _) = self.lstm1(x, (hidden, cells))
        # sum forward and backward nets
        output = output.view(*output.size()[:-1], self.model_dim, 2).sum(dim=-1)
        if hidden is None or cells is None:
            output2, (hidden, cells) = self.lstm2(output)
        else:
            output2, (hidden, cells) = self.lstm2(output, (hidden, cells))
        # sum forward and backward nets
        output2 = output2.view(*output2.size()[:-1], self.model_dim, 2).sum(dim=-1)
        # sum the outputs
        return output + output2, (hidden, cells)


def tile(a, dim, n_tile):
    init_dim = a.size(dim)
    repeat_idx = [1] * a.dim()
    repeat_idx[dim] = n_tile
    a = a.repeat(*(repeat_idx,))
    order_index = torch.LongTensor(torch.cat([init_dim * torch.arange(n_tile) + i for i in range(init_dim)]))
    return torch.index_select(a, dim, order_index.to(a.device))


class Seq2SeqLSTMNetwork(nn.Module):
    device = property(lambda self: next(self.parameters()).device)

    def __init__(self,
                 input_dim: int = 513,
                 model_dim: int = 1024,
                 num_layers: int = 1,
                 n_lstm: int = 1,
                 bottleneck: str = "add",
                 n_fc: int = 1,
                 hop: int = 8,
                 bias: Optional[bool] = False,
                 weight_norm: bool = False,
                 input_module: Optional[nn.Module] = None,
                 output_module: Optional[nn.Module] = None,
                 with_tbptt: bool = False,
                 with_sampler: bool = True
                 ):
        init_ctx = locals()
        super(Seq2SeqLSTMNetwork, self).__init__()
        init_ctx.pop("output_module")
        init_ctx.pop("input_module")
        init_ctx.pop("self")
        init_ctx.pop("__class__")
        self.hp = AttributeDict(init_ctx)
        self.inpt_mod = input_module if input_module is not None else nn.Identity()
        self.enc = EncoderLSTM(input_dim,
                               model_dim, num_layers, n_lstm, bottleneck, n_fc,
                               bias=bias, weight_norm=weight_norm, hop=hop,
                               with_tbptt=with_tbptt)
        self.dec = DecoderLSTM(model_dim, num_layers, bottleneck,
                               bias=bias, weight_norm=(weight_norm,) * 2)
        if with_sampler:
            self.sampler = ParametrizedGaussian(model_dim, model_dim, bias=bias)
        self.outpt_mod = nn.Sequential(
            nn.Linear(model_dim, input_dim, bias=False), Abs()
        ) if output_module is None else output_module
        self.hop = self.rf = self.shift = self.hp.hop
        self.output_length = lambda n: n

    def forward(self, x, temperature=None):
        x = self.inpt_mod(x)
        coded, (h_enc, c_enc) = self.enc(x)
        return self.decode(coded, h_enc, c_enc, temperature)

    def decode(self, x, h_enc, c_enc, temperature=None):
        coded = tile((x.unsqueeze(1) if len(x.shape) < 3 else x), 1, self.hp.hop)
        if self.hp.with_sampler:
            residuals, _, _ = self.sampler(coded)
            coded = coded + residuals
        output, (_, _) = self.dec(coded, h_enc, c_enc)
        return self.outpt_mod(output, *((temperature,) if temperature is not None else ()))

    def reset_hidden(self):
        self.enc.hidden = None

    def before_generate(self, loop, batch, batch_idx):
        self.reset_hidden()
        self.forward(*batch)
        return {}

    def generate_step(self, t, inputs, ctx):
        return self.forward(*inputs)

    def after_generate(self, outputs, ctx, batch_idx):
        self.reset_hidden()


class MultiSeq2SeqLSTM(nn.Module):
    device = property(lambda self: next(self.parameters()).device)

    def __init__(self):
        super(MultiSeq2SeqLSTM, self).__init__()
        self.hp = {}
        lstms = []
        for i in range(3):
            lstms += [Seq2SeqLSTMNetwork(
                input_dim=256 if i > 0 else 513,
                model_dim=256,
                hop=4
            )]
        self.s2s = nn.ModuleList(lstms)

    def forward(self, x, i=0):
        x, (h_enc, c_enc) = self.s2s[i].enc(x)
        if i == len(self.s2s) - 1:
            return self.s2s[i].decode(x, h_enc, c_enc)
        else:
            return self.s2s[i].decode(self.forward(x, i + 1), h_enc, c_enc)

    def reset_hidden(self):
        for s2s in self.s2s:
            s2s.enc.hidden = None

    def generate_step(self, t, inputs, ctx):
        return self.forward(*inputs)



import dataclasses as dtc
import h5mapper as h5m
import json
import os
from .s2s import Seq2SeqLSTM, Seq2SeqLSTMv0
from .wavenets import WaveNetFFT, WaveNetQx
from .srnns import SampleRNN

__all__ = [
    'find_checkpoints',
    'load_trainings_hp',
    'load_network_cls',
    'load_feature',
    'Checkpoint'
]


def find_checkpoints(root="trainings"):
    return h5m.FileWalker(r"\.h5", root)


def load_trainings_hp(dirname):
    return json.loads(open(os.path.join(dirname, "hp.json"), 'r').read())


class CkptBank(h5m.TypedFile):
    ckpt = h5m.TensorDict()


def load_feature(s):
    import mimikit as mmk
    loc = dict()
    exec(f"feature = {s}", mmk.__dict__, loc)
    return loc["feature"]


def load_network_cls(s):
    loc = dict()
    exec(f"cls = {s}", globals(), loc)
    return loc["cls"]


@dtc.dataclass
class Checkpoint:
    id: str
    epoch: int
    root_dir: str = "./"

    @staticmethod
    def get_id_and_epoch(path):
        id_, epoch = path.split("/")[-2:]
        return id_.strip("/"), int(epoch.split(".h5")[0].split("=")[-1])

    @staticmethod
    def from_path(path):
        basename = os.path.dirname(os.path.dirname(path))
        return Checkpoint(*Checkpoint.get_id_and_epoch(path), root_dir=basename)

    @property
    def os_path(self):
        return os.path.join(self.root_dir, f"{self.id}/epoch={self.epoch}.h5")

    def delete(self):
        os.remove(self.os_path)

    @property
    def network(self):
        bank = CkptBank(self.os_path, 'r')
        hp = bank.ckpt.load_hp()
        return bank.ckpt.load_checkpoint(hp["cls"], "state_dict")

    @property
    def feature(self):
        bank = CkptBank(self.os_path, 'r')
        hp = bank.ckpt.load_hp()
        return hp['feature']

    @property
    def train_hp(self):
        return load_trainings_hp(os.path.join(self.root_dir, self.id))



__version__ = '0.3.4'

from . import extract
from . import features
from . import loops
from . import modules
from . import models
from . import networks
from . import demos

from .extract import *
from .features import *
from .loops import *
from .modules import *
from .models import *
from .networks import *

from .train import *
from .utils import *

from .demos import *


def demo():
    """### Load Data"""
    import mimikit as mmk
    import h5mapper as h5m
    import torch
    import os


    # list of files or directories to use as data ("./" is the cwd of the notebook)
    sources = ['./data']
    # audio sample rate
    sr = 16000

    db_path = "train.h5"
    if os.path.exists(db_path):
        os.remove(db_path)

    class SoundBank(h5m.TypedFile):
        snd = h5m.Sound(sr=sr, mono=True, normalize=True)

    SoundBank.create(db_path, sources)
    soundbank = SoundBank(db_path, mode='r', keep_open=True)

    """### Configure and run training"""

    # INPUT / TARGET

    feature = mmk.MuLawSignal(
        sr=soundbank.snd.sr,
        q_levels=256,
    )

    # NETWORK

    net = mmk.models.SampleRNN(
        feature=feature,
        # bottom tier has linearized Mu-Law signal as input (no embeddings!)
        input_type="lin",
        # output MLP learns to predict the temperature of its output distribution
        learn_temperature=True,
        chunk_length=16000 * 8,
        frame_sizes=(256, 128, 64, 32, 16, 8, 4, 4),
        dim=128,
        n_rnn=1,
        q_levels=feature.q_levels,
        embedding_dim=1,
        mlp_dim=128,
    )

    # OPTIMIZATION LOOP

    mmk.train(
        soundbank,
        net,
        input_feature=mmk.MultiScale(feature, net.frame_sizes, (*net.frame_sizes[:-1], 1)),
        target_feature=feature,
        root_dir="./",

        # BATCH

        batch_size=32,
        batch_length=2048,
        oversampling=16,
        shift_error=0,
        tbptt_chunk_length=net.hp.chunk_length,

        # OPTIM

        max_epochs=1000,
        limit_train_batches=None,

        max_lr=1e-3,
        betas=(0.9, 0.9),
        div_factor=1.,
        final_div_factor=10.,
        pct_start=0.05,
        cycle_momentum=False,
        reset_optim=False,

        # MONITORING / OUTPUTS

        CHECKPOINT_TRAINING=True,
        MONITOR_TRAINING=True,
        OUTPUT_TRAINING="",

        every_n_epochs=100,
        n_examples=4,
        # small warm up prompts work better than big ones...
        prompt_length=net.frame_sizes[0]*6,
        n_steps=int(16 * feature.sr),
        temperature=torch.tensor([[1.], [1.5], [.5], [.05]]).repeat(1, int(16*net.feature.sr)),

    )

    """----------------------------"""



import os
from typing import Iterable
from functools import partial
import pytorch_lightning as pl
from matplotlib import pyplot as plt
from pytorch_lightning.trainer.states import TrainerState
from pytorch_lightning import Callback
from IPython import get_ipython

import h5mapper as h5m
from ..utils import audio


__all__ = [
    'is_notebook',
    'EpochProgressBarCallback',
    'GradNormCallback',
    'MMKCheckpoint',
    'GenerateCallback',
    'tqdm'
]


def is_notebook():
    shell = get_ipython().__class__.__name__
    if shell in ('ZMQInteractiveShell', "Shell"):
        # local and colab notebooks
        return True
    elif shell == 'TerminalInteractiveShell':
        return False
    else:
        return False


if is_notebook():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


class EpochProgressBarCallback(Callback):
    def on_train_start(self, trainer, pl_module):
        self.epoch_bar = tqdm(range(1, trainer.max_epochs), unit="epoch",
                              position=0, leave=False, dynamic_ncols=True)

    def on_train_epoch_end(self, trainer, pl_module, unused=None):
        if trainer.val_dataloaders is not None and any(trainer.val_dataloaders):
            return
        else:
            self.epoch_bar.update()

    def on_validation_epoch_end(self, *args):
        self.epoch_bar.update()


class GradNormCallback(Callback):

    def __init__(self):
        self.gradnorms = []

    def on_after_backward(self, trainer, pl_module) -> None:
        self.gradnorms += [pl_module.grad_norm(1.)]


class MMKCheckpoint(Callback):
    """
    checkpoints saved by this Callback can be loaded with :

        net = net_cls(**h5object.attrs)[.with_io(),
        ...].load_state_dict(h5object.get(id_str))

    """
    format = staticmethod(h5m.TensorDict.format)

    def __init__(self,
                 h5_tensor_dict=None,
                 id_template="epoch={epoch}-step={step}",
                 epochs=None,
                 root_dir=''
                 ):
        super().__init__()
        self.h5_tensor_dict = h5_tensor_dict
        self.id_template = id_template
        self.epochs = epochs
        self.root_dir = root_dir
        self.tp_cls = None
        self.hp = {}

    def on_pretrain_routine_start(self, trainer, pl_module) -> None:
        if not self.root_dir:
            self.root_dir = trainer.default_root_dir
        self.hp = pl_module.net.hp
        self.hp.update({"cls": type(pl_module.net)})

        class CkptBank(h5m.TypedFile):
            ckpt = h5m.TensorDict(pl_module.net.state_dict())

        self.tp_cls = CkptBank

    def on_train_epoch_end(self, trainer, pl_module, unused=None) -> None:
        epoch, global_step = trainer.current_epoch + 1, trainer.global_step
        if trainer.state == TrainerState.status.INTERRUPTED or \
                epoch == trainer.max_epochs or self.should_save(epoch, global_step):
            self.save_checkpoint(pl_module, epoch, global_step)

    def should_save(self, epoch, step):
        if type(self.epochs) is int:
            return epoch > 0 and (epoch % self.epochs) == 0
        elif type(self.epochs) is float:  # Doesn't work for now...
            return (step % int(self.epochs)) == 0
        elif isinstance(self.epochs, Iterable):
            return epoch in self.epochs
        return False

    def save_checkpoint(self, pl_module, epoch, global_step):
        new = self.tp_cls(os.path.join(self.root_dir, f"epoch={epoch}.h5"), "w")
        new.ckpt.save_hp(self.hp)
        new.ckpt.add("state_dict", self.format(pl_module.net.state_dict()))
        new.flush()
        new.close()
        # id_str = self.format_id(epoch, global_step)
        # self.h5_tensor_dict.add(id_str, self.format(pl_module.net.state_dict()))
        # self.h5_tensor_dict.flush()

    def format_id(self, epoch, step):
        dct = {"epoch": epoch, "step": step}
        exec(f"out = f'{self.id_template}'", {}, dct)
        return dct["out"]

    def on_fit_end(self, trainer, pl_module) -> None:
        # self.h5_tensor_dict.close()
        pass

    def teardown(self, trainer, pl_module, stage=None) -> None:
        # self.h5_tensor_dict.close()
        pass


class GenerateCallback(pl.callbacks.Callback):

    def __init__(self,
                 generate_loop=None,
                 every_n_epochs=10,
                 plot_audios=True,
                 play_audios=True,
                 output_features=tuple(),
                 audio_logger=None):
        sample_rate = {feat.sr for feat in output_features
                       if getattr(feat, 'sr', False)}
        sample_rate = sample_rate.pop() if len(sample_rate) > 0 else None
        if play_audios and sample_rate is None:
            raise ValueError("Cannot play audio files if no output_feature has a `sr` attribute")
        self.loop = generate_loop
        self.every_n_epochs = every_n_epochs
        self.plot_audios = plot_audios
        self.play_audios = play_audios
        self.output_features = output_features
        self.sample_rate = sample_rate
        self.logger = audio_logger
        self.indices = ()

    def get_prompt_idx(self, batch, output):
        idx = batch * self.loop.dataloader.batch_size + output
        return self.indices[idx]

    def process_outputs(self, outputs, batch_idx, epoch=0):
        for feature, output in zip(self.output_features, outputs):
            output = feature.inverse_transform(output)
            for i, out in enumerate(output):
                y = out.detach().cpu().numpy()
                if self.plot_audios:
                    plt.figure(figsize=(20, 2))
                    plt.plot(y)
                    plt.show(block=False)
                if self.play_audios:
                    audio(y, sr=getattr(feature, 'sr', 22050),
                          hop_length=getattr(feature, 'hop_length', 512))
                self.logger.write(y,
                                  epoch=epoch,
                                  prompt_idx=self.get_prompt_idx(batch_idx, i))

    def on_epoch_end(self, trainer: pl.Trainer, model):
        if (trainer.current_epoch + 1) % self.every_n_epochs != 0:
            return
        self.indices = self.loop.dataloader.sampler.indices
        self.loop.process_outputs = partial(self.process_outputs,
                                            epoch=trainer.current_epoch+1)
        self.loop.run()




import torch
from torch import nn as nn
from typing import Optional
from dataclasses import dataclass

__all__ = [
    'ParametrizedGaussian'
]


@dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class ParametrizedGaussian(nn.Module):
    """
    Parametrized Gaussian (often found in variational auto-encoders)
    """
    input_dim: int
    z_dim: int
    bias: bool = False
    return_params: bool = True

    def __post_init__(self):
        nn.Module.__init__(self)
        self.fc = nn.Linear(self.input_dim, self.z_dim * 2, bias=self.bias)

    def forward(self, h):
        mu, logvar = torch.chunk(self.fc(h), 2, dim=-1)
        std = logvar.mul(0.5).exp_()
        eps = torch.randn(*mu.size()).to(h.device)
        z = mu + std * eps
        if self.return_params:
            return z, mu, std
        return z



import h5mapper as h5m
import torch.nn as nn
import torch
import math
from pprint import pprint

from ..networks import TierNetwork
from ..features import Resample, MuLawSignal
from ..loops import GenerateLoop

from .nnn import NearestNextNeighbor
from .checkpoint import Checkpoint

__all__ = [
    "Ensemble"
]


class Ensemble(nn.Module):
    """
    generate form a prompt by chaining checkpoints/models
    """
    device = property(lambda self: next(self.parameters()).device)

    def __init__(self,
                 max_seconds,
                 base_sr,
                 stream,  # a pbind stream of events
                 print_events=False
                 ):
        super(Ensemble, self).__init__()
        self.max_seconds = max_seconds
        self.base_sr = base_sr
        self.stream = stream
        self.print_events = print_events
        # just to make self.device settable/gettable
        self._param = nn.Parameter(torch.ones(1))

    def run_event(self, inputs, net, feature, n_steps, *params):
        prompt_getter = feature.batch_item(
            shift=0, length=inputs.size(-1) if isinstance(net, TierNetwork) else net.rf  # TODO: NNN needs 1 instead of net.rf...
        )
        resample = Resample(self.base_sr, feature.sr)
        n_input_samples = math.ceil(prompt_getter.getter.length * self.base_sr / feature.sr)
        prompt_getter.data = resample(inputs[:, -n_input_samples:])
        prompt = prompt_getter(0)
        output = [[]]
        print("length", prompt_getter.getter.length, "n_inputs", n_input_samples, "PROMPT", prompt.shape)
        def process_outputs(outputs, _):
            print("OUTPUT", outputs[0].shape)
            inv_resample = Resample(feature.sr, self.base_sr)
            # prompt + generated in base_sr :
            y = feature.inverse_transform(outputs[0])
            print("Y", y.shape, feature.sr, self.base_sr)
            out = inv_resample(y)
            output[0] = out[:, n_input_samples:]
            print("OUTPUT", out.shape)

        loop = GenerateLoop(
            net,
            dataloader=[(prompt,)],
            inputs=(h5m.Input(None,
                              getter=h5m.AsSlice(dim=1, shift=-net.shift, length=net.shift),
                              setter=h5m.Setter(dim=1)),
                    *tuple(h5m.Input(p, h5m.AsSlice(dim=1 + int(hasattr(net.hp, 'hop')), length=1),
                                     setter=None) for p in params)
                    ),
            n_steps=n_steps,
            add_blank=True,
            process_outputs=process_outputs,
            time_hop=getattr(net, 'hop', 1)
        )
        loop.run()
        return output[0]

    @staticmethod
    def seconds_to_n_steps(seconds, net, feature):
        return int(seconds * feature.sr) if isinstance(feature, MuLawSignal) \
            else int(seconds * (feature.sr // feature.hop_length)) // getattr(net, "hop", 1)

    def generate_step(self, t, inputs, ctx):
        if t >= int(self.max_seconds * self.base_sr):
            return None
        event, net, feature, n_steps, params = self.next_event()
        if hasattr(net, 'to'):
            net = net.to("cuda")
        if hasattr(net, "use_fast_generate"):
            net.use_fast_generate = False

        if (t / self.base_sr + event['seconds']) < self.max_seconds:
            if self.print_events:
                event.update({"start": t / self.base_sr})
                pprint(event)
            out = self.run_event(inputs[0], net, feature, n_steps, *params)
            return out
        return torch.zeros(1, int(self.max_seconds * self.base_sr - t)).to("cuda")

    def next_event(self):
        event = next(self.stream)
        if "Checkpoint" in str(event["type"]):
            ck = Checkpoint(event['id'], event['epoch'], event.get("root_dir", "./"))
            net, feature = ck.network, ck.feature
        elif "NearestNextNeighbor" in str(event["type"]):
            feature = event['feature']
            data = event["soundbank"]
            net = NearestNextNeighbor(feature, data.snd)
        else:
            raise TypeError(f"event type '{event['type']}' not recognized")
        n_steps = self.seconds_to_n_steps(event['seconds'], net, feature)
        print("N_STEPS", n_steps)
        if "temperature" in event:
            temp = event['temperature']
            if isinstance(temp, float):
                params = (torch.tensor([[temp]]).to(self.device).repeat(1, n_steps),)
            elif isinstance(temp, tuple):
                params = (torch.linspace(temp[0], temp[1], n_steps, device=self.device),)
            else:
                params = tuple()
        else:
            params = tuple()
        return event, net, feature, n_steps, params



from .callbacks import *
from .generate import *
from .get_trainer import *
from .logger import *
from .train_loops import *
from .samplers import *
from .eval_checkpoint import *

__all__ = [_ for _ in dir() if not _.startswith("_")]





def demo():
    """Generate From Checkpoint"""
    import mimikit as mmk
    import h5mapper as h5m
    import torch
    import matplotlib.pyplot as plt
    import librosa

    # load a checkpoint
    ckpt = mmk.Checkpoint(
        root_dir="./trainings/wn-test-gesten",
        id='84e89798ec2c85e19790344fb598932118c7a65142e747e383907c5f7ced0f26',
        epoch=1
    )
    net, feature = ckpt.network, ckpt.feature

    # prompt positions in seconds
    indices = [
        1.1, 8.5, 46.3
    ]
    # duration in seconds to generate converted to number of steps
    n_steps = librosa.time_to_frames(8, sr=feature.sr, hop_length=feature.hop_length)

    class SoundBank(h5m.TypedFile):
        snd = h5m.Sound(sr=feature.sr, mono=True, normalize=True)

    SoundBank.create("gen.h5", ckpt.train_hp["files"], )
    soundbank = SoundBank("gen.h5", mode='r', keep_open=True)

    def process_outputs(outputs, bidx):
        output = feature.inverse_transform(outputs[0])
        for i, out in enumerate(output):
            y = out.detach().cpu().numpy()
            plt.figure(figsize=(20, 2))
            plt.plot(y)
            plt.show(block=False)
            mmk.audio(y, sr=feature.sr,
                      hop_length=feature.hop_length)

    max_i = soundbank.snd.shape[0] - getattr(feature, "hop_length", 1) * net.rf
    g_dl = soundbank.serve(
        (feature.batch_item(shift=0, length=net.rf, training=False),),
        sampler=mmk.IndicesSampler(N=len(indices),
                                   indices=[librosa.time_to_samples(i, sr=feature.sr) for i in indices],
                                   max_i=max_i,
                                   redraw=False),
        shuffle=False,
        batch_size=len(indices)
    )

    loop = mmk.GenerateLoop(
        network=net,
        dataloader=g_dl,
        inputs=(h5m.Input(None, h5m.AsSlice(dim=1, shift=-net.rf, length=net.rf), setter=h5m.Setter(dim=1)),),
        n_steps=n_steps,
        device='cuda' if torch.cuda.is_available() else 'cpu',
        time_hop=net.hp.get("hop", 1),
        process_outputs=process_outputs
    )
    loop.run()

    """----------------------------"""



import dataclasses
from typing import Optional

import torch
import torch.nn as nn
from dataclasses import dataclass

from pytorch_lightning.utilities import AttributeDict

from .single_class_mlp import SingleClassMLP
from ..modules.homs import *
from .resamplers import *

__all__ = [
    "SampleRNNTier",
    "SampleRNNTopTier",
    "SampleRNNBottomTier",
    "TierNetwork",
    "SampleRNN",
    "MixedRNN",
]


class TierNetwork(HOM):
    device = property(lambda self: next(self.parameters()).device)

    def __init__(self, *tiers):
        in_ = ', '.join([f'x{i}' for i in range(len(tiers))])
        hidden = ', '.join([f'h{i}' for i in range(len(tiers))])
        _, kwargs = zip(*[get_input_signature(tier.forward) for tier in tiers])
        kwargs = set([k for kw in kwargs for k in kw.split(", ") if k != 'h=None'])
        kwargs = ', '.join(kwargs)
        super().__init__(
            f"{in_}, {kwargs} -> z",
            (lambda self: self.hidden, f"self -> {hidden}"),
            *[
                (tier, ', '.join(tier.s.in_).replace("x", f"x{i}").replace("h", f"h{i}") + f" -> z, h{i}")
                for i, tier in enumerate(tiers)
            ],
            (lambda *args: setattr(args[0], 'hidden', args[1:]), f"self, {hidden} -> $")
        )
        self.tiers = tiers
        self.frame_sizes = tuple(tier.hp.frame_size for tier in tiers)
        self.hidden = (None,) * len(tiers)
        self.shift = self.rf = self.frame_sizes[0]
        self.outputs = []
        self.rf = self.shift = self.frame_sizes[0]
        self.output_length = lambda n: n

    def reset_hidden(self):
        self.hidden = (None,) * len(self.frame_sizes)

    def before_generate(self, loop, batch, batch_idx):
        self.outputs = [None] * (len(self.frame_sizes) - 1)
        assert batch[0].size(1) % self.shift == 0, f'batch_length must be divisible by {self.shift} ' \
                                                   f'(got {batch[0].size(1)})'
        self.reset_hidden()
        self.hidden = list(self.hidden)
        self.prior_t = len(batch[0][0])
        # warm-up
        for t in range(self.shift, self.prior_t):
            self.generate_step(t, (batch[0][:, t - self.shift:t],), {})
        return {}

    def generate_step(self, t, inputs, ctx):
        tiers = self.tiers
        hiddens = self.hidden
        outputs = self.outputs
        fs = self.frame_sizes
        # only one tensor in the `inputs` tuple
        temperature = inputs[1] if len(inputs) > 1 else None
        inputs = inputs[0]
        for i in range(len(tiers) - 1):
            if t % fs[i] == 0:
                inpt = inputs[:, -fs[i]:].unsqueeze(1)
                if i == 0:
                    prev_out = None
                else:
                    prev_out = outputs[i - 1][:, (t // fs[i]) % (fs[i - 1] // fs[i])].unsqueeze(1)
                out, h = tiers[i](inpt, prev_out, hiddens[i])
                hiddens[i] = h
                outputs[i] = out
        if t < self.prior_t:
            return
        prev_out = outputs[-1]
        inpt = inputs[:, -fs[-1]:].reshape(-1, 1, fs[-1])
        out, _ = tiers[-1](inpt, prev_out[:, (t % fs[-2]) - fs[-2]].unsqueeze(1), temperature=temperature)
        return out.squeeze(-1) if len(out.size()) > 2 else out

    def after_generate(self, *args, **kwargs):
        self.outputs = []
        self.reset_hidden()


class SampleRNNTier(HOM):
    resampler_cls = None
    hp = AttributeDict()

    device = property(lambda self: next(self.parameters()).device)

    def linearize(self, qx):
        """ maps input samples (0 <= qx < 256) to floats (-2. <= x < 2.) """
        return ((qx.float() / self.hp.q_levels) - .5) * 4

    @staticmethod
    def add_upper_tier(x, upper):
        return x if upper is None else (x + upper)

    def reset_hidden(self, x, h):
        if h is None or x.size(0) != h[0].size(1):
            B = x.size(0)
            h0 = nn.Parameter(torch.randn(self.hp.n_rnn, B, self.hp.dim).to(x.device))
            c0 = nn.Parameter(torch.randn(self.hp.n_rnn, B, self.hp.dim).to(x.device))
            return h0, c0
        else:
            return tuple(h_.detach() for h_ in h)


class SampleRNNTopTier(SampleRNNTier):

    def __init__(self,
                 frame_size: int,
                 dim: int,
                 up_sampling: int = 1,
                 n_rnn: int = 2,
                 q_levels: int = 256,
                 linearize=True,
                 ):
        init_ctx = locals()
        init_ctx.pop("self")
        init_ctx.pop("__class__")
        self.hp = AttributeDict(init_ctx)
        self.frame_size = FS = frame_size
        super().__init__(
            "x, z=None, h=None -> x, h",
            *Maybe(linearize,
                   (self.linearize, "x -> x")),
            (nn.Linear(FS, dim), "x -> x"),
            (self.add_upper_tier, "x, z -> x"),
            (self.reset_hidden, "x, h -> h"),
            (nn.LSTM(dim, dim, n_rnn, batch_first=True), "x, h -> x, h"),
            (LinearResampler(dim, t_factor=up_sampling, d_factor=1), "x -> x"),
        )


class SampleRNNBottomTier(SampleRNNTier):

    def __init__(self,
                 frame_size: int,
                 top_dim: int,
                 zin_dim: int,
                 zout_dim: int,
                 io_dim: Optional[int] = None,
                 input_type: str = "lin",
                 learn_temperature: bool = True,
                 n_hidden_layers: int = 0,
                 ):
        init_ctx = locals()
        init_ctx.pop("self")
        init_ctx.pop("__class__")
        self.hp = AttributeDict(init_ctx)
        self.hp.q_levels = io_dim
        self.frame_size = frame_size
        super().__init__(
            f"x, z=None, h=None, temperature=None -> x, h",
            *Maybe(io_dim is not None,
                   *Maybe(input_type == "emb",
                          (nn.Embedding(io_dim, zin_dim), 'x -> x')),
                   *Maybe(input_type == "lin",
                          (lambda x: self.linearize(x.unsqueeze(-1)), "x -> x"))
                   ),
            (Conv1dResampler(zin_dim, 1 / frame_size, top_dim / zin_dim), 'x -> x'),
            (self.add_upper_tier, "x, z -> x"),
            *Maybe(io_dim is not None,
                   (SingleClassMLP(top_dim, zout_dim, io_dim,
                                   learn_temperature=learn_temperature,
                                   n_hidden_layers=n_hidden_layers), 'x, temperature -> x'))
        )


@dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class SampleRNN(TierNetwork):
    frame_sizes: tuple = (16, 8, 8)  # from top to bottom!
    dim: int = 512
    n_rnn: int = 2
    q_levels: int = 256
    embedding_dim: int = 256
    mlp_dim: int = 512
    input_type: str = "lin"  # bottom tier input type ["emb" | "lin"]
    learn_temperature: bool = True
    n_hidden_layers: int = 0

    def __post_init__(self):
        self.hp = AttributeDict(dataclasses.asdict(self))
        tiers = []
        for i, fs in enumerate(self.frame_sizes[:-1]):
            tiers += [SampleRNNTopTier(fs, self.dim,
                                       up_sampling=fs // (
                                           self.frame_sizes[i + 1]
                                           if i < len(self.frame_sizes) - 2
                                           else 1),
                                       n_rnn=self.n_rnn,
                                       q_levels=self.q_levels,
                                       )]
        tiers += [SampleRNNBottomTier(self.frame_sizes[-1], self.dim,
                                      zin_dim=self.embedding_dim,
                                      zout_dim=self.mlp_dim,
                                      io_dim=self.q_levels,
                                      input_type=self.input_type,
                                      learn_temperature=self.learn_temperature,
                                      n_hidden_layers=self.n_hidden_layers,
        )]
        TierNetwork.__init__(self, *tiers)


@dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class MixedRNN(TierNetwork):
    """top tiers get mag-fft inputs, bottom tier get embedded samples"""

    frame_sizes: tuple = (128, 64, 8)  # from top to bottom!
    dim: int = 512
    n_rnn: int = 2
    q_levels: int = 256
    embedding_dim: int = 256
    mlp_dim: int = 512

    def __post_init__(self):
        self.hp = AttributeDict(dataclasses.asdict(self))
        tiers = []
        for i, fs in enumerate(self.frame_sizes[:-1]):
            tiers += [SampleRNNTopTier(fs, self.dim,
                                       up_sampling=fs // (self.frame_sizes[i + 1]
                                                          if fs != self.frame_sizes[i + 1] else 1),
                                       n_rnn=self.n_rnn,
                                       linearize=False,
                                       )]
        tiers += [SampleRNNBottomTier(self.frame_sizes[-1], self.dim,
                                      zin_dim=self.embedding_dim,
                                      zout_dim=self.mlp_dim,
                                      io_dim=self.q_levels,
                                      )]
        TierNetwork.__init__(self, *tiers)
